/********************************************************************
	write_idblock.c


	Writes product ID Block(s) to top of flash and verifies.
	This version modified for Dynamic C 7.04P2 and later.

	This program is intended to write a System_ID Block on boards
	that don't have a System_ID block or that have a version 2
	system_ID block. If your board has a version 3 or 4 system_ID
	block then you should use the idblock_recovery.c utility program
	for proper recovery of your data from the userblock data area.

	This program will automatically prompt you if it detects
	a version 3 or 4 system_ID block to see if you want to
	continue with writing the system_ID block.  

	Version 1.8

 ********************************************************************


 Instructions:


 1. Uncomment the MYBOARD macro below for the board you are writing
    to, or add your own MYBOARD macro for a custom board.


 2. Enter the MAC address of the board you are writing to if it is
    Ethernet capable.  For Z-World products, the first three hex
    digits are always 0x00, 0x90, 0xC2.  The last three digits should
    be on a sticker on your product.  NOTE that OP6700 and Rabbit
    TCP/IP Development Kit boards already have MAC addresses in an
    onboard EEPROM and versions 1.1 or later of this program will not
    write a MAC address to the ID block even if supplied.


3.  Define the NUMUSER4KBLOCKS macro to 0 for no User Block, or to
    the number of 4K blocks in the desired User Block size.
    EG:  Define NUMUSER4KBLOCKS to 2 for an 8K User Block size.


4.  Run the program.  The contents of the ID block will be printed
    (or any errors reported) to stdio.


 ********************************************************************


	History:
			1.8
				- Added error messages when compiling as primary (DLM)
				  or secondary (DLP) type program, and when compiling
				  with separate I&D space enabled.
				- Automatically mirror tableVersion 3 and up ID blocks
				  as required.
				- Replaced HASUSERBLOCK macro with NUMUSER4KBLOCKS macro
				  to facilitate installing a custom User Block size.
				- Removed RCM2130 from the HAS2FLASHES macro list.
				- Added boards:  BL2010B, OP6805, OP72xx, RCM21xx,
				  RCM2210, RCM34xx, SR9160.
				- Reorganized and renamed Has2Flash(), HasEEPROM(), and
				  HasEthernet() macros to HAS2FLASHES, HASEEPROM, and
				  HASETHERNET macro lists respectively, to make
				  maintenance easier when adding new board types.
				- Standardized board descriptions.
			1.7
				- Made "unconditional" use of div19200ptr.
			1.6
				- Added board:  LP3510.
				- Added RCM3200 to HasEthernet() macro.
				- Forces version 2 ID Block for Dynamic C 7.25 and earlier.
				- Added detection of system ID blocks of version 3 or
				  higher to prompt the user to use the idblock_recovery
				  program.
			1.5
				- Separated ifndef board type groups into individual board
				  ID macros.
				- Conditionally defined GCPU, GREV register addresses.
			1.4
				- Added more boards:  BL21xx, LP3500, and OP72xx.
				- Bug with reading divider19200 fixed in Dynamic C 7.30;
				  needed to modify routine to do it the right way.
				- Made ID block version dependent on Dynamic C version
				  program is compiled with:
						7.06 or earlier will write a type 2 ID block
						7.20 or later will write type 4
			1.3
				- Added Rabbit 3000 cores (RCM3xxx series).
			1.2
				- Updated for version 4 ID block (new fields).
				- Improved support for older Dynamic C versions.


			1.1
				- Fixed a small bug in getFlashID().
				- Added RCM3000 core modules.
				- Fixed problem with writing MAC address in ID block to
				  OP6700 and RTDKs, which would override the MAC address
				  in the on-board EEPROM.
			1.0
				- Corresponds to initwrid.c version 3.14.


 ********************************************************************/
			
#define VERSION "1.8"
#define PROGRAM_NAME "write_idblock.c"


/********************************************************************/
/**  STEP ONE:  uncomment the macro for the board you are using.   **/
/********************************************************************/


//*** RabbitCore Modules ***//

//#define	MYBOARD RCM2000	// 25MHz R2000, 256K flash, 512k SRAM
//#define	MYBOARD RCM2010	// 25MHz R2000, 256K flash, 128k SRAM
//#define	MYBOARD RCM2020	// 18MHz R2000, 256K flash, 128K SRAM

//#define	MYBOARD RCM2100	// 22MHz R2000, 2x256K flash, 512K SRAM, ethernet
//#define	MYBOARD RCM2110	// 22MHz R2000, 256K flash, 128K SRAM, ethernet
//#define	MYBOARD RCM2115	// 22MHz R2000, 256K flash, 128K SRAM, ethernet (no RJ45)
//#define	MYBOARD RCM2120	// 22MHz R2000, 2x256K flash, 512K SRAM
//#define	MYBOARD RCM2130	// 22MHz R2000, 256K flash, 128K SRAM

//#define	MYBOARD RCM2200	// 22MHz R2000, 256K flash, 128K SRAM, ethernet
//#define	MYBOARD RCM2210	// 22MHz R2000, 256K flash, 128K SRAM, ethernet (no RJ45)
//#define	MYBOARD RCM2250	// 22MHz R2000, 2x256K flash, 512K SRAM, ethernet

//#define	MYBOARD RCM2300	// 22MHz R2000, 256K flash, 128K SRAM

//#define	MYBOARD RCM3000	// 29MHz R3000, 2x256K flash, 512K SRAM, ethernet
//#define	MYBOARD RCM3010	// 29MHz R3000, 256K flash, 128K SRAM, ethernet

//#define	MYBOARD RCM3100	// 29MHz R3000, 2x256K flash, 512K SRAM
//#define	MYBOARD RCM3110	// 29MHz R3000, 256K flash, 128K SRAM

//#define	MYBOARD RCM3200	// 44MHz R3000, 512K flash, 256K+512K SRAM, ethernet

//#define	MYBOARD RCM3400A	// 29MHz R3000, 512K flash, 512K SRAM, ethernet on PB, analog
//#define	MYBOARD RCM3410A	// 29MHz R3000, 256K flash, 256K SRAM, ethernet on PB, analog


//*** Single Board Computers ***//

//Jackrabbit
//#define	MYBOARD BL1800		// 29MHz R2000, 256K flash, 128K SRAM, RS485, battery
//#define	MYBOARD BL1805		// 29MHz R2000, 256K flash, 512K SRAM, RS485, battery
//#define	MYBOARD BL1810		// 14MHz R2000, 128K flash, 128K SRAM, RS485, battery
//#define	MYBOARD BL1820		// 14MHz R2000, 128K flash, 128k SRAM

//Wildcat
//#define	MYBOARD BL2000		// 22MHz R2000, 256K flash, 128K SRAM, ethernet, high-quality analog
//#define	MYBOARD BL2010		// 22MHz R2000, 256K flash, 128K SRAM, ethernet, analog
//#define	MYBOARD BL2010B	// 22MHz R2000, 256K flash, 512K SRAM, ethernet, analog
//#define	MYBOARD BL2020		// 22MHz R2000, 256K flash, 128K SRAM, high-quality analog
//#define	MYBOARD BL2030		// 22MHz R2000, 256K flash, 128K SRAM, analog
//#define	MYBOARD BL2040		// 22MHz R2000, 256K flash, 128K SRAM

//Smart Cat
//#define	MYBOARD BL2100		// 22MHz RCM2200, 256K flash, 128K SRAM, ethernet, analog
//#define	MYBOARD BL2101		// 22MHz RCM2200, 256K flash, 128K SRAM, ethernet, 10V analog
//#define	MYBOARD BL2105		// 22MHz RCM2250, 2x256K flash, 512K SRAM, ethernet, analog
//#define	MYBOARD BL2110		// 22MHz RCM2200, 256K flash, 128K SRAM, ethernet
//#define	MYBOARD BL2111		// 22MHz RCM2250, 2x256K flash, 512K SRAM, ethernet, 10V analog
//#define	MYBOARD BL2115		// 22MHz RCM2250, 2x256K flash, 512K SRAM, ethernet
//#define	MYBOARD BL2120		// 22MHz RCM2300, 256K flash, 128K SRAM, analog
//#define	MYBOARD BL2121		// 22MHz RCM2300, 256K flash, 128K SRAM, 10V analog
//#define	MYBOARD BL2130		// 22MHz RCM2300, 256K flash, 128K SRAM

//Fox Low-Power Controller
//#define	MYBOARD LP3500		// 7MHz R3000, 2x256K flash, 512K SRAM, analog, relay
//#define	MYBOARD LP3510		// 7MHz R3000, 256K flash, 128K SRAM


//*** Embedded Control Systems ***//

//SmartStar
//#define	MYBOARD SR9100		// 29MHz R2000, 2x256K flash, 128K SRAM

//MiniStar
//#define	MYBOARD SR9150		// 22MHz R2000, 2x256K flash, 128K SRAM, ethernet
//#define	MYBOARD SR9160		// 22MHz R2000, 2x256K flash, 128K SRAM

//*** Operator Interfaces ***//

//Intellicom
//#define	MYBOARD OP6600		// 18MHz R2000, 256K flash, 128K SRAM
//#define	MYBOARD OP6700		// 18MHz R2000, 2x256K flash, 256K SRAM, ethernet

//Minicom
//#define	MYBOARD OP6800		// 22MHz RCM2200, 256K flash, 128K SRAM, ethernet
//#define	MYBOARD OP6805		// 22MHz RCM2250, 2x256K flash, 512K SRAM, ethernet
//#define	MYBOARD OP6810		// 22MHz RCM2300, 256K flash, 128K SRAM

//eDisplay
//#define	MYBOARD OP7200		// 22MHz RCM2200, 256K flash, 128K SRAM, ethernet, analog, touchscreen
//#define	MYBOARD OP7200B	// 22MHZ RCM2250, 2x256K flash, 512K SRAM, ethernet, analog, touchscreen
//#define	MYBOARD OP7210		// 22MHz RCM2300, 256K flash, 128K SRAM
//#define	MYBOARD OP7210B	// 22MHz RCM2250, 2x256K flash, 512K SRAM, ethernet


//*** Other Products ***//

//RabbitLink
//#define	MYBOARD RABLINK	// 22MHz R2000, 2x256K flash, 128K SRAM, ethernet

// TCP/IP Development Kit
//#define	MYBOARD RTDK		// 18MHz R2000, 2x256K flash, 128K SRAM, ethernet


/********************************************************************/
/**  STEP TWO:  enter the MAC address of the board if it is	       **/
/**		Ethernet-capable.  the first three digits are fixed; the  **/
/**		last three should be on a sticker on the board.           **/
/********************************************************************/


const char MAC[6] = { 0x00, 0x90, 0xC2,
							 0x00, 0x00, 0x00 };


/********************************************************************/
/**  STEP THREE:  set the NUMUSER4KBLOCKS macro to the desired     **/
/**		number of 4K blocks in the User Block.  if 0, only the    **/
/**      ID block will be written, and no User Block will be       **/
/**      installed.  eg:  set to 2 to install an 8K User Block.    **/
/********************************************************************/


#define	 NUMUSER4KBLOCKS	4


/********************************************************************/
/**  END OF USER-CUSTOMIZABLE SECTION				                   **/
/********************************************************************/


// check for undefined register macros
//  (R2000C and up has these, prior R2000s always return 0s)


#ifndef GCPU
#define GCPU 0x2E
#endif


#ifndef GREV
#define GREV 0x2F
#endif


// check for several potential problems


#ifndef _FLASH_
#error "This program _must_ be run in flash."
#endif


#ifdef __SEPARATE_INST_DATA__
#if __SEPARATE_INST_DATA__
#error "This program, as written, is not compatible with separate I&D space."
#endif
#endif


#ifdef COMPILE_SECONDARY_PROG
#error "This utility is not intended for use as a secondary (DLP) program."
#error "Must remove COMPILE_SECONDARY_PROG macro from Compiler Options Defines box."
#error "Must also type Ctrl-Y to recompile the BIOS."
#endif


#ifdef COMPILE_PRIMARY_PROG
#error "This utility is not intended for use as a primary (DLM) program."
#error "Must remove COMPILE_PRIMARY_PROG macro from Compiler Options' Defines box."
#error "Must also type Ctrl-Y to recompile the BIOS."
#endif


#ifndef MYBOARD
#error "MYBOARD must be set to the proper board type."
#endif


#if ((NUMUSER4KBLOCKS < 0) || (NUMUSER4KBLOCKS > 6))
#error "NUMUSER4KBLOCKS must be set from 0 to 6 (0K to 24K), inclusive."
#endif


// check for undefined board types
//  (new since 7.04P2)


#ifndef BL2000
	#define BL2000		0x0800
#endif
#ifndef BL2010
	#define BL2010		0x0801
#endif
#ifndef BL2010B
	#define BL2010B	0x0805
#endif
#ifndef BL2020
	#define BL2020		0x0802
#endif
#ifndef BL2030
	#define BL2030		0x0803
#endif
#ifndef BL2040
	#define BL2040		0x0804
#endif
#ifndef BL2100
	#define BL2100		0x0B00
#endif
#ifndef BL2101
	#define BL2101		0x0B06
#endif
#ifndef BL2105
	#define BL2105		0x0B04
#endif
#ifndef BL2110
	#define BL2110		0x0B01
#endif
#ifndef BL2111
	#define BL2111		0x0B08
#endif
#ifndef BL2115
	#define BL2115		0x0B05
#endif
#ifndef BL2120
	#define BL2120		0x0B02
#endif
#ifndef BL2121
	#define BL2121		0x0B07
#endif
#ifndef BL2130
	#define BL2130		0x0B03
#endif
#ifndef LP3500
	#define LP3500		0x1200
#endif
#ifndef LP3510
	#define LP3510		0x1201
#endif
#ifndef OP6600
	#define OP6600		0x0300
#endif
#ifndef OP6700
	#define OP6700		0x0302
#endif
#ifndef OP6800
	#define OP6800		0x0D00
#endif
#ifndef OP6805
	#define OP6805		0x0D02
#endif
#ifndef OP6810
	#define OP6810		0x0D01
#endif
#ifndef OP7200
	#define OP7200		0x1100
#endif
#ifndef OP7200B
	#define OP7200B	0x1102
#endif
#ifndef OP7210
	#define OP7210		0x1101
#endif
#ifndef OP7210B
	#define OP7210B	0x1103
#endif
#ifndef RABLINK
	#define RABLINK	0x0601
#endif
#ifndef RCM2100
	#define RCM2100	0x0700
#endif
#ifndef RCM2110
	#define RCM2110	0x0701
#endif
#ifndef RCM2115
	#define RCM2115	0x0704
#endif
#ifndef RCM2120
	#define RCM2120	0x0702
#endif
#ifndef RCM2130
	#define RCM2130	0x0703
#endif
#ifndef RCM2200
	#define RCM2200	0x0900
#endif
#ifndef RCM2210
	#define RCM2210	RCM2200
#endif
#ifndef RCM2250
	#define RCM2250	0x0E00
#endif
#ifndef RCM3000
	#define RCM3000	0x0F00
#endif
#ifndef RCM3010
	#define RCM3010	0x0F01
#endif
#ifndef RCM3100
	#define RCM3100	0x1000
#endif
#ifndef RCM3110
	#define RCM3110	0x1001
#endif
#ifndef RCM3200
	#define RCM3200	0x1300
#endif
#ifndef RCM3400A
	#define RCM3400A	0x1600
#endif
#ifndef RCM3410A
	#define RCM3410A	0x1601
#endif
#ifndef RTDK
	#define RTDK		0x0301
#endif
#ifndef SR9100
	#define SR9100		0x0502
#endif
#ifndef SR9150
	#define SR9150		0x0C00
#endif
#ifndef SR9160
	#define SR9160		0x0C01
#endif


// some useful macros


#if	((MYBOARD == BL2105) || (MYBOARD == BL2111) || (MYBOARD == BL2115))
	#define HAS2FLASHES 1
#elif	(MYBOARD == LP3500)
	#define HAS2FLASHES 1
#elif	(MYBOARD == OP6600)
	#define HAS2FLASHES 1
#elif	(MYBOARD == OP6700)
	#define HAS2FLASHES 1
#elif	(MYBOARD == OP6805)
	#define HAS2FLASHES 1
#elif	((MYBOARD == OP7200B) || (MYBOARD == OP7210B))
	#define HAS2FLASHES 1
#elif	(MYBOARD == RABLINK)
	#define HAS2FLASHES 1
#elif	((MYBOARD == RCM2100) || (MYBOARD == RCM2120))
	#define HAS2FLASHES 1
#elif	(MYBOARD == RCM2250)
	#define HAS2FLASHES 1
#elif	(MYBOARD == RCM3000)
	#define HAS2FLASHES 1
#elif	(MYBOARD == RCM3100)
	#define HAS2FLASHES 1
#elif	(MYBOARD == RTDK)
	#define HAS2FLASHES 1
#elif	((MYBOARD == SR9100) || (MYBOARD == SR9150) || (MYBOARD == SR9160))
	#define HAS2FLASHES 1
#else
	#define HAS2FLASHES 0
#endif


#if	(MYBOARD == OP6700)
	#define HASEEPROM 1
#elif	(MYBOARD == RTDK)
	#define HASEEPROM 1
#else
	#define HASEEPROM 0
#endif


#if	((MYBOARD == BL2000) || (MYBOARD == BL2010) || (MYBOARD == BL2010B))
	#define HASETHERNET 1
#elif	((MYBOARD == BL2100) || (MYBOARD == BL2101) || (MYBOARD == BL2105))
	#define HASETHERNET 1
#elif	((MYBOARD == BL2110) || (MYBOARD == BL2111) || (MYBOARD == BL2115))
	#define HASETHERNET 1
#elif	(MYBOARD == OP6700)
	#define HASETHERNET 1
#elif	((MYBOARD == OP6800) || (MYBOARD == OP6805))
	#define HASETHERNET 1
#elif	((MYBOARD == OP7200) || (MYBOARD == OP7200B))
	#define HASETHERNET 1
#elif	(MYBOARD == RABLINK)
	#define HASETHERNET 1
#elif	((MYBOARD == RCM2100) || (MYBOARD == RCM2110) || (MYBOARD == RCM2115))
	#define HASETHERNET 1
#elif	((MYBOARD == RCM2200) || (MYBOARD == RCM2210) || (MYBOARD == RCM2250))
	#define HASETHERNET 1
#elif	((MYBOARD == RCM3000) || (MYBOARD == RCM3010))
	#define HASETHERNET 1
#elif	(MYBOARD == RCM3200)
	#define HASETHERNET 1
#elif	(MYBOARD == RTDK)
	#define HASETHERNET 1
#elif	(MYBOARD == SR9150)
	#define HASETHERNET 1
#else
	#define HASETHERNET 0
#endif


// function prototypes


int writeIDBlock();
int getFlashID();
int calcCRC(int initCRC, void *data, int blocksize);
int setTimeStamp();


/********************************************************************/


main()
{
	static int	i, crc;
	int	saveMB1;
	char	*div19200ptr;
	char ch;
	
	if(SysIDBlock.tableVersion > 2 && CC_VER < 0x0726)
	{
		printf("\n\n");
		printf("Detected a version %d SystemID block\n", SysIDBlock.tableVersion);
		printf("Recommend the %s utility for proper Userblock data recovery.\n", PROGRAM_NAME);
		printf("(i.e. Z-World calibration constants or User application code data)\n\n");
		printf("Do you still want to overwrite your SystemID block (Y/N)? ") ;
		ch = getchar();
		if(ch == 'N' || ch == 'n')
		{
			printf("\n\n");
			printf("Exiting without making any changes...\n\n");
			exit(0);
		}
		printf("\n\n");
	}
	
	
	printf("WRITE_IDBLOCK, version %s\n", VERSION);


	//////////////////////////////////////////////////
	// map first two quadrants to flash on /CS0
	// make sure A18 NOT inverted on MB1 (handles bank select)


	saveMB1 = MB1CRShadow;
	WrPortI(MB1CR, &MB1CRShadow, FLASH_WSTATES | 0x00);
	_InitFlashDriver(0x01 | 0x02);


	//////////////////////////////////////////////////
	// set up the ID block fields


	memset(&SysIDBlock, 0x00, sizeof(SysIDBlock));

	#if (CC_VER < 0x0726)
	SysIDBlock.tableVersion = 2;
	#else
	SysIDBlock.tableVersion = 4;
	#endif

	// Uncomment the following assignment to force a version 2
	//  ID Block, regardless of the Dynamic C compiler version.
	//	SysIDBlock.tableVersion = 2;


#if (NUMUSER4KBLOCKS & 1)
	if (SysIDBlock.tableVersion >= 3) {
		printf("\n\nCurrently, the BIOS only supports mirrored User Block (ID\n");
		printf("Block version 3 and up) sizes that are a multiple of 8KB.\n\n");
		printf("Recommend either editing the source to force a version 2 ID\n");
		printf("Block, or define NUMUSER4KBLOCKS to be an even number.\n\n");
		printf("Exiting without making any changes...\n\n");
		exit(0);
	}
#endif


	SysIDBlock.vendorID = 1;
	setTimeStamp();

	SysIDBlock.idBlockSize = sizeof(SysIDBlock);

	for (i=0; i<3; i++) {
		SysIDBlock.marker[i*2]   = 0x55;
		SysIDBlock.marker[i*2+1] = 0xAA;
	}

	
	//////////////////////////////////////////////////
	// board-specific data


	SysIDBlock.productID = MYBOARD;
	SysIDBlock.flashID = getFlashID();

	SysIDBlock.flashSize = _FLASH_SIZE_;
	SysIDBlock.flashType = _FlashInfo.writeMode;
	SysIDBlock.sectorSize = _FlashInfo.sectorSize;
	SysIDBlock.numSectors = _FlashInfo.numSectors;
	SysIDBlock.flashSpeed = 0x00;   // special case, means "unknown"
	
#if HAS2FLASHES
	// we assume that the 2nd flash (if installed) is
	//  identical to the first one
	SysIDBlock.flash2ID = SysIDBlock.flashID;
	SysIDBlock.flash2Size = _FLASH_SIZE_;
	SysIDBlock.flash2Type = _FlashInfo.writeMode;
	SysIDBlock.sector2Size = _FlashInfo.sectorSize;
	SysIDBlock.num2Sectors = _FlashInfo.numSectors;
	SysIDBlock.flash2Speed = 0x00;  // special case, means "unknown"
#endif

	SysIDBlock.ramID = 0x01;
	SysIDBlock.ramSize = _RAM_SIZE_;
	SysIDBlock.ramSpeed = 0x00;	// special case, means "unknown"

	SysIDBlock.cpuID = getCPUID();	  // Rabbit CPU + Version
	div19200ptr = divider19200;
	SysIDBlock.crystalFreq = (*div19200ptr) * 614400ul;


	// may fill serial number and product name here (future)


#if NUMUSER4KBLOCKS
	// reserve space for the User Block
	SysIDBlock.userBlockSize =
	SysIDBlock.userBlockLoc = (NUMUSER4KBLOCKS * 4096)
	                          - (int)SysIDBlock.idBlockSize;
#endif

	
#if (HASETHERNET && !HASEEPROM)
	// fill MAC address
	memcpy(SysIDBlock.macAddr, MAC, 6);
#endif

				
	//////////////////////////////////////////////////
	// calculate the ID block's CRC sum


	SysIDBlock.idBlockCRC = 0x0000;

	i = (int)(&SysIDBlock.reserved) - (int)(&SysIDBlock.tableVersion);
	crc = calcCRC(0x0000, &SysIDBlock, i);
	crc = calcCRC(crc, &SysIDBlock.idBlockSize, 16);

	SysIDBlock.idBlockCRC = crc;

				
	//////////////////////////////////////////////////
	// write the ID Block(s) to flash


	i = writeIDBlock();
	if (i != 0)	{
		printf("\nError writing ID block (%d): \n", i);
		if (i == -1)	printf("  attempt to write to non-flash area\n");
		else if (i == -2)	printf("  source not located in root\n");
		else if (i == -3)	printf("  timeout during flash write\n");
		else printf("  unknown error type\n");
		exit(1);
	} else {
		printf("\nID block successfully written.\n\n");
	}


	//////////////////////////////////////////////////
	// now try and read the ID block back...


	memset(&SysIDBlock, 0x00, sizeof(SysIDBlock));
	i = _readIDBlock(0x01 | 0x02);

	if (i != 0) {
		printf("Error reading ID block (%d)\n", i);
		exit(1);
	} else {
		printf("ID block read successfully:\n");
		printf("  Version	= %d\n", SysIDBlock.tableVersion);
		printf("  ID block size = %d bytes\n", SysIDBlock.idBlockSize);
		printf("  Product ID    = 0x%04X\n", SysIDBlock.productID);
		printf("  Vendor ID	= %d\n", SysIDBlock.vendorID);
		printf("  Timestamp	= %02d/%02d/%02d%02d  %02d:%02d:%02d\n",
			SysIDBlock.timestamp[2], SysIDBlock.timestamp[3],
			SysIDBlock.timestamp[0], SysIDBlock.timestamp[1],
			SysIDBlock.timestamp[4], SysIDBlock.timestamp[5],
			SysIDBlock.timestamp[6]);

		printf("\n");
		printf("  CPU	   = Rabbit ");
		if (SysIDBlock.cpuID < 0x0100)		  printf("2000\n");
		else if (SysIDBlock.cpuID == 0x0100)	printf("3000\n");
		else												printf("????\n");
		printf("  Crystal Freq  = %8.4f MHz\n", SysIDBlock.crystalFreq/1.0e6);
		printf("  Serial number = '%s'\n", SysIDBlock.serialNumber);
		printf("  Product name  = '%s'\n", SysIDBlock.productName);
		printf("  MAC address   = %02X:%02X:%02X:%02X:%02X:%02X\n",
						SysIDBlock.macAddr[0], SysIDBlock.macAddr[1],
						SysIDBlock.macAddr[2], SysIDBlock.macAddr[3],
						SysIDBlock.macAddr[4], SysIDBlock.macAddr[5]);
		
		printf("\n");		   
		printf("  User block offset = %04x\n", SysIDBlock.userBlockLoc);
		printf("  User block size   = %04x\n", SysIDBlock.userBlockSize);

		printf("\n");
		printf("  Flash 1 information:\n");
		printf("    ID	  = 0x%04X\n", SysIDBlock.flashID);
		printf("    Type	= %d\n", SysIDBlock.flashType);
		printf("    Size	= %ld\n", SysIDBlock.flashSize * 4096ul);
		printf("    Sector size = %d bytes\n", SysIDBlock.sectorSize);
		printf("    Num sectors = %d\n", SysIDBlock.numSectors);

		printf("\n");
		if (SysIDBlock.flash2ID == 0x0000) {
			printf("  No second flash\n");
		} else {
			printf("  Flash 2 information:\n");
			printf("    ID	  = 0x%04X\n", SysIDBlock.flash2ID);
			printf("    Type	= %d\n", SysIDBlock.flash2Type);
			printf("    Size	= %ld\n", SysIDBlock.flash2Size * 4096ul);
			printf("    Sector size = %d bytes\n", SysIDBlock.sector2Size);
			printf("    Num sectors = %d\n", SysIDBlock.num2Sectors);
		}

		printf("\n");
		printf("  RAM Information:\n");
		printf("    ID		= 0x%04X\n", SysIDBlock.ramID);
		printf("    Size	   = %ld bytes\n", SysIDBlock.ramSize * 4096ul);
	}


	//////////////////////////////////////////////////
	// restore memory quadrant 1


	WrPortI(MB1CR, &MB1CRShadow, saveMB1);
}


/********************************************************************/


int writeIDBlock()
{
	static unsigned long	physaddr;
	static int	addr, blockSize;
	static char	xpc;
	static int	err;

	blockSize = sizeof(SysIDBlock);
	physaddr = 0x00080000ul - blockSize;

	addr = 0xE000 + (int)(physaddr & 0x0FFFul);
	xpc  = (int)(((physaddr - (unsigned long)addr) & 0xFF000ul) >> 12ul);

	_overwrite_block_flag = 1;
	err = WriteFlash(physaddr, &SysIDBlock, blockSize);
	_overwrite_block_flag = 0;

#if NUMUSER4KBLOCKS
	// Has a User Block, now check if mirrored . . .
	if (SysIDBlock.tableVersion >= 3) {
		// Mirrored ID/User Block version, might as well do it now!
		physaddr -= NUMUSER4KBLOCKS * 4096L;
		_overwrite_block_flag = 1;
		err = WriteFlash(physaddr, &SysIDBlock, blockSize);
		_overwrite_block_flag = 0;
	}
#endif

	return(err);
}


/********************************************************************/


int getFlashID()
{
	static int	id;

#asm
	call	_GetFlashID
	ld		(id), hl
#endasm

	return(id);
}


/********************************************************************/


int getCPUID()
{
	static int	id;

#asm
	ioi	ld a, (GCPU)
	and	0x1F
	ld		(id+1), a
	ioi	ld a, (GREV)
	and	0x1F
	ld		(id), a
#endasm

	return(id);
}


/********************************************************************/


int setTimeStamp()
{
	static struct tm	t;

	tm_rd(&t);
	SysIDBlock.timestamp[0] = (t.tm_year + 1900) / 100;
	SysIDBlock.timestamp[1] = t.tm_year % 100;
	SysIDBlock.timestamp[2] = t.tm_mon;
	SysIDBlock.timestamp[3] = t.tm_mday;
	SysIDBlock.timestamp[4] = t.tm_hour;
	SysIDBlock.timestamp[5] = t.tm_min;
	SysIDBlock.timestamp[6] = t.tm_sec;
}


/********************************************************************/


#define CRC_POLY		0x1021	  // 16-bit CRC polynomial
											//		(recommended by CCITT X25 standard)


/*
 *	CRC calculation functions
 *		- can be called from C (calcCRC) or assembly (_calcCRC)
 *		- reference:  "A painless guide to CRC error detection algorithms",
 *							8/19/1993, Ross N. Williams
 */


#asm
//	int calcCRC(int initCRC, void *data, int blocksize);
calcCRC::
	ld		hl, 2
	add	hl, sp
	ld		hl, (hl)
	ex		de, hl			  ; initial CRC value

	ld		hl, 6				   ; the "blocksize" variable
	add	hl, sp
	ld		hl, (hl)
	ld		b, h
	ld		c, l					; get data size off stack


	ld		hl, 4				   ; the "data" variable
	add	hl, sp
	ld		hl, (hl)				; hl now contains pointer to data


;; assembly entry point
_calcCRC::
;		de contains initial CRC value
;		hl contains pointer to data
;		bc contains number of bytes
dataloop:
	push	hl
	ld		h, (hl)			 ; get next byte into hl (and shift it left 8 bits)
	ld		l, 0x00

	push	bc						; save byte counter
	call	_byteCRC				; call CRC function from BIOS
	pop	bc						; restore byte counter

	pop	hl
	inc	hl
	dec	bc

	xor	a
	cp		b
	jr		nz, dataloop
	cp		c
	jr		nz, dataloop

	ex		de, hl			  ; CRC returned in hl
	ret
#endasm


//////////////////////////////////////////////////////////////////////


#asm
;;
;;	Calculate the CRC value for a single byte.  Can be called multiple
;;	times (without resetting the CRC value passed to it) to do a CRC
;;	check on a larger block of data.
;;
;;		data byte passed in h (l should be 0x00)
;;		crc value passed, returned in de
;;
_byteCRC::
	push	bc
	push	af

	ld		b, 0x08			 ; bit counter
crcloop:
	push	bc						; save bit counter
	ld		a, h
	xor	d
	scf
	ccf							; clear carry bit
	rl		de						; roll crc left one bit
	ex		de, hl
	scf
	ccf							; clear carry bit
	rl		de						; roll data left one bit
	ex		de, hl
	bit	7, a					; result from earlier XOR
	jr		z, bit7iszero
bit7isone:
	ld		bc, CRC_POLY		; recommended 16-bit polynomial (X.25)
	ld		a, e
	xor	c
	ld		e, a
	ld		a, d					; XOR crc with polynomial
	xor	b
	ld		d, a
bit7iszero:
	pop	bc						; restore bit counter
	djnz	crcloop

	pop	af
	pop	bc
	ret
#endasm
